# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from typing import Any

from libinephany.observations.observation_utils import StatisticStorageTypes
from libinephany.observations.observers.base_observers import GlobalObserver
from libinephany.observations.observers.global_observers.base_classes import LHOPTCheckpointBaseObserver
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class TrainingProgress(GlobalObserver):

    @property
    def can_standardize(self) -> bool:
        """
        :return: Whether the observation can be standardized.
        """

        return False

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        return observation_inputs.training_progress

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class EpochsCompleted(GlobalObserver):

    @property
    def can_standardize(self) -> bool:
        """
        :return: Whether the observation can be standardized.
        """

        return False

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        return observation_inputs.epochs_completed

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class ProgressAtEachCheckpoint(LHOPTCheckpointBaseObserver):
    """
    This is a global observer from the paper "Learning to Optimize with Reinforcement Learning"
    https://arxiv.org/abs/2305.18291.

    It returns a single float value that is the training progress at the current checkpoint.
    The observation tracks training progress and returns the progress value only when a checkpoint is reached.
    """

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float:
        """
        Returns training progress at each checkpoint interval.

        Before checkpoint: returns progress towards next checkpoint (0 to 1)
        At checkpoint: returns the actual training progress value
        """
        current_progress = observation_inputs.training_progress

        # Cold start: If the last progress is not set, set it to the first progress record
        self._cold_start(current_progress)

        self._update_history(current_progress)

        # Check if we should create a checkpoint
        if self._should_create_checkpoint():
            # Return the progress at this checkpoint
            self.last_value = current_progress
            return current_progress
        else:
            return self.last_value



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from typing import TypeAlias

import torch.nn as nn

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

DropoutLayer: TypeAlias = (
    nn.Dropout | nn.Dropout1d | nn.Dropout2d | nn.Dropout3d | nn.AlphaDropout | nn.FeatureAlphaDropout
)

# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================


def create_torch_dropout_mapping(model: nn.Module, parameter_group_names: list[str]) -> dict[str, list[DropoutLayer]]:
    """
    :param model: Model to create dropout mapping from.
    :param parameter_group_names: Names of the parameter groups.
    :return: Dictionary mapping parameter group names to a list of the dropout layers present in that parameter group.
    """

    dropout_mapping: dict[str, list[DropoutLayer]] = {
        parameter_group_name: [] for parameter_group_name in parameter_group_names
    }

    for parameter_group_name in parameter_group_names:
        for module_name, module in model.named_modules():
            if parameter_group_name not in module_name:
                continue

            if isinstance(module, DropoutLayer):
                dropout_mapping[parameter_group_name].append(module)

    return dropout_mapping


def set_torch_dropout(
    dropout_layer_mapping: dict[str, list[DropoutLayer]] | None, parameter_group_name: str, dropout: float
) -> None:
    """
    :param dropout_layer_mapping: Mapping from parameter group names to a list of the dropout layers present in that
    parameter group.
    :param parameter_group_name: Name of the parameter group to set the dropout probability of.
    :param dropout: Dropout to set the given parameter group to.

    :note: Operation done in-place.
    """

    if dropout_layer_mapping is None:
        raise ValueError("No dropout layer mapping present. Cannot get dropout.")

    if parameter_group_name not in dropout_layer_mapping:
        return

    for dropout_layer in dropout_layer_mapping[parameter_group_name]:
        dropout_layer.p = dropout


def get_torch_parameter_group_dropout(
    dropout_layer_mapping: dict[str, list[DropoutLayer]] | None, parameter_group_name: str
) -> float | None:
    """
    :param dropout_layer_mapping: Mapping from parameter group names to a list of the dropout layers present in that
    parameter group.
    :param parameter_group_name: Name of the parameter group to get the dropout probability of.

    :note: Operation done in-place.
    """

    if dropout_layer_mapping is None:
        raise ValueError("No dropout layer mapping present. Cannot get dropout.")

    if parameter_group_name not in dropout_layer_mapping:
        return None

    if dropout_layer_mapping[parameter_group_name]:
        return dropout_layer_mapping[parameter_group_name][0].p

    return None



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import random
import secrets

import numpy as np
import torch

# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================


def generate_true_random_seed(num_bits: int = 30) -> int:
    """
    :param num_bits: Number of bits to use in the random seed.
    :return: Truly random seed.
    """

    return secrets.randbits(k=num_bits)


def generate_random_seed(num_bits: int = 32, seed: int | None = None) -> int:
    """
    :param num_bits: Number of bits to use in the random seed.
    :param seed: Random seed used to generate the random seed in order to maintain consistency between developers.
    :return: Truly random seed.
    """

    if seed is not None:
        random.seed(seed)

    return random.getrandbits(num_bits)


def set_all_seeds(seed: int) -> None:
    """
    :param seed: Random seed to use for random, numpy, torch, CUDA and CUDNN.
    """

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)

    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)

    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from abc import ABC, abstractmethod
from enum import Enum
from typing import Callable

import numpy as np

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class HyperparameterTransformType(Enum):

    IDENTITY = "identity"
    LOG = "log"
    LOG1M = "log1m"


class HyperparameterTransform(ABC):
    """Abstract class for hyperparameter transformations."""

    @abstractmethod
    def to_internal(self, value: float | int) -> float | int:
        """Transforms an 'external' hyperparameter value into an 'internal' one."""

        raise NotImplementedError

    @abstractmethod
    def to_external(self, value: float | int) -> float | int:
        """Transforms an 'internal' hyperparameter value into an 'external' one."""

        raise NotImplementedError


class IdentityTransform(HyperparameterTransform):
    """Applies an identity transform to a hyperparameter."""

    def to_internal(self, value: float | int) -> float | int:
        return value

    def to_external(self, value: float | int) -> float | int:
        return value


class LogTransform(HyperparameterTransform):
    """Applies a logarithmic transform to a hyperparameter."""

    def to_internal(self, value: float | int) -> float | int:
        return np.log(value)

    def to_external(self, value: float | int) -> float | int:
        return np.exp(value)


class Log1mTransform(HyperparameterTransform):
    """Transforms a variable by log(1-x)."""

    def to_internal(self, value: float | int) -> float | int:
        return np.log1p(-value)

    def to_external(self, value: float | int) -> float | int:
        return -np.expm1(value)


# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

HYPERPARAMETER_TRANSFORM_REGISTRY: dict[HyperparameterTransformType, type[HyperparameterTransform]] = {
    HyperparameterTransformType.IDENTITY: IdentityTransform,
    HyperparameterTransformType.LOG: LogTransform,
    HyperparameterTransformType.LOG1M: Log1mTransform,
}

# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================


def transform_and_sort_hparam_values(
    transform_fn: Callable[[float | int], float | int], first_value: float | int, second_value: float | int
) -> tuple[float | int, float | int]:
    """
    :param transform_fn: The transform function to apply to the values.
    :param first_value: First value to sort.
    :param second_value: Second value to sort.
    :return: Tuple of the two given values with the smaller value being the first element.
    """
    first_value = transform_fn(first_value)
    second_value = transform_fn(second_value)

    return min(first_value, second_value), max(first_value, second_value)



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import math
from typing import Any

from libinephany.observations import observation_utils, statistic_trackers
from libinephany.observations.observation_utils import StatisticStorageTypes
from libinephany.observations.observers.base_observers import GlobalObserver
from libinephany.observations.observers.global_observers.base_classes import LHOPTBaseObserver
from libinephany.observations.observers.global_observers.constants import LHOPT_CONSTANTS
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class GlobalFirstOrderGradients(GlobalObserver):

    @property
    def requires_include_statistics(self) -> bool:
        """
        :return: Whether the observation requires include_statistics to be provided.
        """

        return True

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.TENSOR_STATISTICS

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        statistics = tracked_statistics[statistic_trackers.FirstOrderGradients.__name__]

        return observation_utils.average_tensor_statistics(tensor_statistics=list(statistics.values()))  # type: ignore

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {statistic_trackers.FirstOrderGradients.__name__: dict(include_statistics=self.include_statistics)}


class GlobalSecondOrderGradients(GlobalObserver):

    def __init__(
        self,
        *,
        compute_hessian_diagonal: bool = False,
        **kwargs,
    ) -> None:
        """
        :param compute_hessian_diagonal: Whether to compute the Hessian diagonal to determine second order gradients
        or use the squared first order gradients as approximations in the same way Adam does.
        :param kwargs: Miscellaneous keyword arguments.
        """

        super().__init__(**kwargs)

        self.compute_hessian_diagonal = compute_hessian_diagonal

    @property
    def requires_include_statistics(self) -> bool:
        """
        :return: Whether the observation requires include_statistics to be provided.
        """

        return True

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.TENSOR_STATISTICS

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        statistics = tracked_statistics[statistic_trackers.SecondOrderGradients.__name__]

        return observation_utils.average_tensor_statistics(tensor_statistics=list(statistics.values()))  # type: ignore

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {
            statistic_trackers.SecondOrderGradients.__name__: dict(
                include_statistics=self.include_statistics, compute_hessian_diagonal=self.compute_hessian_diagonal
            )
        }


class LHOPTGradientVarianceFraction(LHOPTBaseObserver):
    """
    This is a global observer from the OpenAI paper "Learning to Optimize with Reinforcement Learning"
    https://arxiv.org/abs/2305.18291.

    It returns two-dimensional observations: [raw_value, cdf_feature] for gradient variance fraction values.
    """

    def __init__(
        self,
        *,
        variance_threshold: float = LHOPT_CONSTANTS["DEFAULT_VARIANCE_THRESHOLD"],
        **kwargs,
    ) -> None:
        """
        :param variance_threshold: Threshold for variance comparison in gradient variance fraction calculation
        :param kwargs: Other observation keyword arguments.
        """
        super().__init__(**kwargs)
        self.variance_threshold = variance_threshold

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """
        return 2  # [raw_value, cdf_feature]

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        raw_value = list(tracked_statistics[statistic_trackers.GradientVarianceFraction.__name__].values())[0]  # type: ignore[list-item]

        cdf_feature = self._compute_cdf_feature(raw_value)  # type: ignore[arg-type]
        self._update_time()

        return [raw_value, cdf_feature]  # type: ignore[list-item]

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {
            statistic_trackers.GradientVarianceFraction.__name__: dict(
                variance_threshold=self.variance_threshold, sample_frequency=LHOPT_CONSTANTS["DEFAULT_SAMPLE_FREQUENCY"]
            ),
        }


class LHOPTMomentumGradientRatio(LHOPTBaseObserver):
    """
    This is a global observer from the OpenAI paper "Learning to Optimize with Reinforcement Learning"
    https://arxiv.org/abs/2305.18291.

    It returns two-dimensional observations: [raw_value, cdf_feature] for momentum gradient ratio values.
    """

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """
        return 2  # [raw_value, cdf_feature]

    @property
    def requires_include_statistics(self) -> bool:
        """
        :return: Whether the observation requires include_statistics to be provided.
        """

        return True

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        statistics = tracked_statistics[statistic_trackers.MomentumGradientRatioStatistics.__name__]

        raw_value = list(statistics.values())[0]  # type: ignore[list-item]

        cdf_feature = self._compute_cdf_feature(raw_value)  # type: ignore[arg-type]
        self._update_time()

        return [raw_value, cdf_feature]  # type: ignore[list-item]

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {
            statistic_trackers.MomentumGradientRatioStatistics.__name__: dict(
                include_statistics=self.include_statistics,
                sample_frequency=LHOPT_CONSTANTS["DEFAULT_SAMPLE_FREQUENCY"],
            ),
        }


class CosineSimilarityObserverOfGradientAndMomentum(LHOPTBaseObserver):
    """
    This is a global observer from the OpenAI paper "Learning to Optimize with Reinforcement Learning"
    https://arxiv.org/abs/2305.18291.

    It returns two-dimensional observations: [raw_value, cdf_feature] for cosine similarity of gradient and momentum values.
    """

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """
        return 3  # [raw_value, cdf_feature, logit_of_cdf_feature]

    @property
    def requires_include_statistics(self) -> bool:
        """
        :return: Whether the observation requires include_statistics to be provided.
        """

        return True

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        statistics = tracked_statistics[
            statistic_trackers.CosineSimilarityObserverOfGradientAndMomentumStatistics.__name__
        ]

        raw_value = list(statistics.values())[0]  # type: ignore[list-item]

        cdf_feature = self._compute_cdf_feature(raw_value)  # type: ignore[arg-type]
        self._update_time()

        # Handle edge cases for logit calculation
        if cdf_feature <= 0.0 or cdf_feature >= 1.0:
            logit_of_cdf_feature = 0.0
        else:
            logit_of_cdf_feature = math.log(cdf_feature / (1 - cdf_feature))

        return [raw_value, cdf_feature, logit_of_cdf_feature]  # type: ignore[list-item]

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {
            statistic_trackers.CosineSimilarityObserverOfGradientAndMomentumStatistics.__name__: dict(
                include_statistics=self.include_statistics,
                sample_frequency=LHOPT_CONSTANTS["DEFAULT_SAMPLE_FREQUENCY"],
            )
        }


class CosineSimilarityObserverOfGradientAndUpdate(LHOPTBaseObserver):
    """
    This is a global observer from the OpenAI paper "Learning to Optimize with Reinforcement Learning"
    https://arxiv.org/abs/2305.18291.

    It returns two-dimensional observations: [raw_value, cdf_feature] for cosine similarity of gradient and update values.
    """

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """
        return 3  # [raw_value, cdf_feature, logit_of_cdf_feature]

    @property
    def requires_include_statistics(self) -> bool:
        """
        :return: Whether the observation requires include_statistics to be provided.
        """

        return True

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        statistics = tracked_statistics[
            statistic_trackers.CosineSimilarityObserverOfGradientAndUpdateStatistics.__name__
        ]

        raw_value = list(statistics.values())[0]  # type: ignore[list-item]

        cdf_feature = self._compute_cdf_feature(raw_value)  # type: ignore[arg-type]
        self._update_time()

        # Handle edge cases for logit calculation
        if cdf_feature <= 0.0 or cdf_feature >= 1.0:
            logit_of_cdf_feature = 0.0
        else:
            logit_of_cdf_feature = math.log(cdf_feature / (1 - cdf_feature))

        return [raw_value, cdf_feature, logit_of_cdf_feature]  # type: ignore[list-item]

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {
            statistic_trackers.CosineSimilarityObserverOfGradientAndUpdateStatistics.__name__: dict(
                include_statistics=self.include_statistics,
                sample_frequency=LHOPT_CONSTANTS["DEFAULT_SAMPLE_FREQUENCY"],
            )
        }


class CosineSimilarityOfGradientAndParameter(LHOPTBaseObserver):
    """
    This is a global observer from the OpenAI paper "Learning to Optimize with Reinforcement Learning"
    https://arxiv.org/abs/2305.18291.

    It returns two-dimensional observations: [raw_value, cdf_feature] for cosine similarity of gradient and parameter values.
    """

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """
        return 3  # [raw_value, cdf_feature, logit_of_cdf_feature]

    @property
    def requires_include_statistics(self) -> bool:
        """
        :return: Whether the observation requires include_statistics to be provided.
        """

        return True

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        statistics = tracked_statistics[statistic_trackers.CosineSimilarityOfGradientAndParameterStatistics.__name__]

        raw_value = list(statistics.values())[0]  # type: ignore[list-item]

        cdf_feature = self._compute_cdf_feature(raw_value)  # type: ignore[arg-type]
        self._update_time()

        # Handle edge cases for logit calculation
        if cdf_feature <= 0.0 or cdf_feature >= 1.0:
            logit_of_cdf_feature = 0.0
        else:
            logit_of_cdf_feature = math.log(cdf_feature / (1 - cdf_feature))

        return [raw_value, cdf_feature, logit_of_cdf_feature]  # type: ignore[list-item]

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {
            statistic_trackers.CosineSimilarityOfGradientAndParameterStatistics.__name__: dict(
                include_statistics=self.include_statistics,
                sample_frequency=LHOPT_CONSTANTS["DEFAULT_SAMPLE_FREQUENCY"],
            )
        }



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from typing import Any

import numpy as np
from numpy.typing import DTypeLike
from pydantic import BaseModel, model_validator

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class ObservationInputs(BaseModel):
    training_loss: float
    validation_loss: float

    training_score: float
    validation_score: float

    training_progress: float
    epochs_completed: int

    @model_validator(mode="before")
    def none_to_zero(cls, data: dict[str, Any]) -> dict[str, Any]:
        """
        :param data: Data to validate and turn all fields given None as their value who have float or int annotations
        to 0.0.
        :return: Validated model data.
        """

        for field, value in data.items():
            field_type = cls.model_fields[field].annotation

            if value is None and (field_type is int or field_type is float):
                data[field] = 0.0 if field_type is float else 0

        return data


class Observations(BaseModel):

    observation_inputs: ObservationInputs

    hit_invalid_value: bool

    agent_observations: dict[str, list[float | int]]
    observations_as_dict: dict[str, dict[str, list[float | int] | dict[str, float]]] | None = None

    def observations_as_arrays(self, dtype: DTypeLike = np.float32) -> dict[str, np.ndarray]:
        """
        :param dtype: Data type to cast the observations to.
        :return: Dictionary mapping agent IDs to their observation vectors as numpy arrays.
        """

        return {
            agent_id: np.array(observation_vector, dtype=dtype)
            for agent_id, observation_vector in self.agent_observations.items()
        }



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from collections import defaultdict
from typing import Any, Callable, DefaultDict

import torch
import torch.nn as nn
import torch.optim as optim

from libinephany.observations import observation_utils, statistic_trackers
from libinephany.observations.statistic_trackers import Statistic
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.utils import torch_distributed_utils
from libinephany.utils.constants import PARAMS, SCHEDULER_GROUP_NAME

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class StatisticManager:

    def __init__(
        self,
        model: nn.Module,
        optimizer: optim.Optimizer,
        agent_modules: dict[str, str],
        can_nullify_gradients: bool = True,
        max_statistic_cache_size: int = 3,
        tensor_stats_downsample_percentage: float = 0.01,
        statistic_sample_frequency: int = 10,
    ) -> None:
        """
        :param model: Model to register forward hook trackers with.
        :param optimizer: Inner model optimizer which contains the parameter groups various metrics can be computed
        from.
        :param agent_modules: Dictionary mapping agent controlled module names to the type of module being controlled.
        :param can_nullify_gradients: Whether the manager can nullify the gradients after computing the various
        statistics. Clients can disable this if they need the gradients for further processing.
        :param max_statistic_cache_size: Maximum number of tensors to store internally before processing the tensors
        into statistics.
        :param tensor_stats_downsample_percentage: Percentage of elements of the tensors from the model to randomly
        sample to compute the statistics from.
        :param statistic_sample_frequency: How frequently to cache tensors from the model for later statistic
        computation.
        """

        self._required_trackers: dict[str, dict[str, Any] | None] | None = None

        self._trackers: dict[str, Statistic] = {}
        self.averaging_function = observation_utils.get_exponential_weighted_average  # Hard-coded for now.

        self.model = model
        self.optimizer = optimizer
        self.agent_modules = agent_modules
        self.can_nullify_gradients = can_nullify_gradients
        self.max_statistic_cache_size = max_statistic_cache_size
        self.tensor_stats_downsample_percentage = tensor_stats_downsample_percentage
        self.statistic_sample_frequency = statistic_sample_frequency

    @property
    def requires_gradient_graphs(self) -> bool:
        """
        :return: Whether the manager requires gradient graphs to be retained.
        """

        return any(tracker.requires_gradient_graphs for tracker in self._trackers.values())

    @staticmethod
    def _nullify_gradients(model: nn.Module) -> None:
        """
        :param model: Model to nullify the gradient of.
        """

        for parameter in model.parameters():
            parameter.grad = None

    @staticmethod
    def _clip_gradients(
        optimizer: optim.Optimizer,
        clipping_thresholds: DefaultDict[str, float],
        clipping_function: Callable[[list[torch.Tensor], float], None],
    ) -> None:
        """
        :param optimizer: Inner model optimizer which contains the parameter groups various metrics can be computed
        from.
        :param clipping_thresholds: Dictionary mapping parameter group names to gradient clipping thresholds for that
        parameter group.
        :param clipping_function: Function used to clip the gradients.
        """

        for parameter_group in optimizer.param_groups:
            parameter_group_name = parameter_group[SCHEDULER_GROUP_NAME]

            if parameter_group_name not in clipping_thresholds:
                continue

            clipping_threshold = clipping_thresholds[parameter_group_name]

            if clipping_threshold is None:
                continue

            clipping_function(parameter_group[PARAMS], clipping_threshold)

    def _iterate_trackers(
        self,
        optimizer: optim.Optimizer,
        model: nn.Module,
    ) -> None:
        """
        :param optimizer: Inner model optimizer which contains the parameter groups various metrics can be computed
        from.
        :param model: Inner model to gather statistics from.
        """

        tracker_order = None
        if torch_distributed_utils.is_scheduler_master_rank():
            tracker_order = list(self._trackers.keys())

        tracker_order = torch_distributed_utils.broadcast_data(data=tracker_order)

        for tracker_name in tracker_order:
            tracker = self._trackers[tracker_name]
            tracker.gather(optimizer=optimizer, model=model)

    def on_optimizer_step(
        self,
        *,
        clipping_function: Callable[[list[torch.Tensor], float], None] | None,
        clipping_thresholds: DefaultDict[str, float] | None = None,
    ) -> None:
        """
        :param clipping_function: Function used to clip the gradients.
        :param clipping_thresholds: Dictionary mapping parameter group names to gradient clipping thresholds for that
        parameter group.
        """

        if clipping_thresholds is not None and clipping_function is not None:
            self._clip_gradients(
                optimizer=self.optimizer, clipping_thresholds=clipping_thresholds, clipping_function=clipping_function
            )

        self._iterate_trackers(optimizer=self.optimizer, model=self.model)

        if self.can_nullify_gradients:
            self._nullify_gradients(model=self.model)

    def compile(self) -> dict[str, dict[str, float | TensorStatistics]] | None:
        """
        :return: Dictionary mapping statistic tracker class names to dictionaries mapping module names to floats or
        TensorStatistic models or None if the local rank is not the master rank.
        """

        if not torch_distributed_utils.is_scheduler_master_rank():
            return None

        compiled: DefaultDict[str, dict[str, float | TensorStatistics]] = defaultdict(dict)
        for tracker in self._trackers.values():
            data = tracker.fetch()
            compiled[tracker.__class__.__name__][tracker.parameter_group_name] = data

        return dict(compiled)

    def build_trackers(
        self,
        required_trackers: dict[str, dict[str, Any] | None] | None,
    ) -> None:
        """
        :param required_trackers: Dictionary mapping statistic tracker class names to kwargs for that class.
        """

        if required_trackers is not None:
            self._required_trackers = required_trackers

        if self._required_trackers is None:
            raise ValueError(
                f"{self.__class__.__name__} must be called with required_trackers not None at least once before trackers "
                f"can be constructed implicitly."
            )

        for required_tracker, tracker_kwargs in self._required_trackers.items():
            try:
                tracker_class: type[Statistic] = getattr(statistic_trackers, required_tracker)

            except AttributeError as e:
                raise AttributeError(f"Statistic tracker class {required_tracker} was not recognised!") from e

            if tracker_kwargs is None:
                tracker_kwargs = {}

            for parameter_group in self.optimizer.param_groups:
                if SCHEDULER_GROUP_NAME not in parameter_group:
                    continue

                parameter_group_name = parameter_group[SCHEDULER_GROUP_NAME]

                tracker = tracker_class(
                    parameter_group_name=parameter_group_name,
                    averaging_function=self.averaging_function,
                    agent_modules=self.agent_modules,
                    max_statistic_cache_size=self.max_statistic_cache_size,
                    tensor_stats_downsample_percentage=self.tensor_stats_downsample_percentage,
                    statistic_sample_frequency=self.statistic_sample_frequency,
                    **tracker_kwargs,
                )

                self._trackers[tracker.tracker_name] = tracker

                if tracker.uses_forward_hook:
                    tracker.register(model=self.model)

    def reset(self) -> None:
        """
        Resets internal state and rebuilds trackers for new model.
        """

        self._trackers = {}

        self.build_trackers(required_trackers=None)
